#ifndef KITCHENSINK_H
#define KITCHENSINK_H

/**
 * @brief Header aggregator
 * 
 * @file kitchensink.h
 * @author Tuomas Virtanen
 * @date 2018-06-27
 */

#include "kitchensink/kitlib.h"
#include "kitchensink/kiterror.h"
#include "kitchensink/kitformat.h"
#include "kitchensink/kitcodec.h"
#include "kitchensink/kitsource.h"
#include "kitchensink/kitplayer.h"
#include "kitchensink/kitutils.h"
#include "kitchensink/kitconfig.h"

#endif // KITCHENSINK_H
